/*jshint expr:true eqnull:true */
/**
 *
 * Backbone.DeepModel v{{version}}
 *
 * Copyright (c) 2013 Charles Davison, Pow Media Ltd
 *
 * https://github.com/powmedia/backbone-deep-model
 * Licensed under the MIT License
 */

{{body}}
